import json
import os
import socket
import subprocess
from threading import Thread
from time import time, sleep
import requests

yafu_running = False
yafu_started_at = time()
yafu_test_seconds = 31

# settings for GPU filtering
gpu_node_ip = os.environ["GPU_NODE"]
#gpu_node_ip = "192.168.178.43:4000"
cado_node = "127.0.0.1:7777"
yafu_debug = False
use_yafu_filter = True
is_gpu_filter = True

# settings for CPU filtering
#gpu_node_ip = "192.168.178.59:4000"
#cado_node = "127.0.0.1:7777"
#yafu_debug = False
#use_yafu_filter = False
#is_gpu_filter = False

candidate = ""
current_block = -1
candidate_level = 1
candidate_block = 0
candidate_segment = 1

pool_settings = {}
last_loaded_at = -1

VERSION = "3.1"

def is_block_ended():
    try:
        x = requests.get('http://' + gpu_node_ip + "/currentblock")
        block_height = x.text
        global current_block
        if current_block == -1:
            current_block = block_height
            return False

        if current_block < block_height:
            current_block = block_height
            return True
    except:
        print("no connection to master node - cannot determine block end")
        return False

def startYafu():
    global yafu_running
    global yafu_started_at
    global candidate

    if yafu_running:
        return

    if candidate == "":
        return

    yafu_running = True
    yafu_started_at = time()

    print("next candidate ", candidate)
    p2 = subprocess.run(["sh", "yafurun.sh", str(candidate)], capture_output=True, encoding="utf-8")
    candidate = ""

    if yafu_debug is True:
        print(p2.stdout)
        print(p2.stderr)

    print("yafu ended")

    yafu_running = False
    yafu_started_at = -1

def checkRuntime():
    global yafu_running
    global yafu_started_at
    global candidate
    global yafu_test_seconds
    global current_block
    global candidate_level
    global candidate_block
    global candidate_segment

    while True:
        if yafu_running == False:
            sleep(1)
            continue

        if is_block_ended() and yafu_running:
            print()
            print("block " + current_block + " ended. stopping yafu...")
            candidate = ""
            subprocess.run(["killall", "yafu"])
            subprocess.run(["killall", "yafu512"])
            yafu_running = False
            continue

        duration = time() - yafu_started_at

        print("yafu running since: ", duration, "  sec. ", end="\r", flush=True)

        if duration > yafu_test_seconds:
            try:
                print()
                print()
                print("Submitting good candidate: ", candidate)
                r = requests.post('http://' + cado_node + '/submitcandidate', json={"cand": candidate, "level": candidate_level, "block": candidate_block, "segment": candidate_segment})
                candidate = ""
                subprocess.run(["killall", "yafu"])
                subprocess.run(["killall", "yafu512"])
                yafu_running = False
            except:
                print("failed to submit candidate. is cado miner running?")
                sleep(1)

        sleep(0.5)

def getPoolSettings():
    global pool_settings
    global last_loaded_at

    if last_loaded_at == -1 or time() - last_loaded_at > 60:
        last_loaded_at = time()
        # load settings from pool
        try:
            x = requests.get('https://fact.zapto.org/requirements')
            data = x.json()
            pool_settings = data
            return data
        except:
            print("could not load settings from pool. using defaults ...")

    return pool_settings

def mine():
    global yafu_running
    global candidate
    global candidate_level
    global current_block
    global candidate_block
    global VERSION
    global use_yafu_filter
    global is_gpu_filter
    global candidate_segment

    print("Starting miner_node version ", VERSION, " - gpu host: ", gpu_node_ip)

    threadRuntime = Thread(target=checkRuntime)
    threadRuntime.start()

    # good
    #candidate = "2570818561336167978058582648390094258117670727613773335664884439741116089917910861255574954478622647756777986592583"

    # bad
    # candidate = "90377629292003121684002147101760858109247336549001090677693"

    while True:
        if yafu_running:
            sleep(1)
            continue

        try:
            is_block_ended()
            x = requests.get('http://' + cado_node + "/needcandidate")
            resp = x.text
        except:
            print("no connection to fact-miner")
            sleep(2)
            continue

        #if resp == "false":
            #print ("Block: ", current_block, " - cado doesnt need candidates...")

        if resp == "true" and candidate != "":
            print("cado wants a candidate but current candidate is not empty")

        # cado wants a candidate
        if resp == "true" and candidate == "":
            try:
                hostname = socket.gethostname()
                x = requests.get('http://' + gpu_node_ip + "/getcandidate?hostname=" + hostname)
                if x.text == "":
                    print("no candidate received. waiting..")
                    sleep(1)
                    continue
            except:
                print("no connection to gpu node")
                sleep(1)
                continue

            try:
                print("-->", x)

                segment = 1

                data = x.json()
                gpu_cand = data["cand"]
                level = data["level"]
                block = data["block"]

                candidate_block = block

                if "segment" in data:
                    segment = data["segment"]

                print("time: ", time(), " - received candidate:", data)
                pool_settings = getPoolSettings()

                use_multi_segment_mode = False
                if 'multiSegmentMode' in pool_settings:
                    use_multi_segment_mode = pool_settings['multiSegmentMode']

                if (use_yafu_filter == False or segment == 2) and not use_multi_segment_mode:
                    level = 3

                # level 1 candidate from gpu filter returned -> will have to rerun yafu on that
                if is_gpu_filter and level == 1:
                    candidate = gpu_cand
                    candidate_segment = segment
                    candidate_block = block
                    thread = Thread(target=startYafu)
                    thread.start()
                    sleep(1)
                    continue

                if gpu_cand != "" and (level == 3 or is_gpu_filter):
                    print("Submitting level 3 (no yafu run) candidate: ", candidate)
                    r = requests.post('http://' + cado_node + '/submitcandidate', json={"cand": gpu_cand, "level": level, "block": block, "segment": segment})
                    candidate = ""
                    sleep(3)
                    continue

                if gpu_cand != "":
                    candidate = gpu_cand
                    candidate_level = data["level"]
                    candidate_block = block
                    candidate_segment = segment
            except Exception as e:
                print("failed fetching candidate. will continue.... . error was: ", str(e))

        thread = Thread(target=startYafu)
        thread.start()

        sleep(1)



if __name__ == "__main__":
    mine()